#include "mediator.h"
#include "worldmodel.h"
#include "perceptron.h"
#include "skill.h"
#include "log.h"
#include "agent.h"

/**************   PrioredCommand *************************/
PrioredCommand::PrioredCommand(){
	Command::Command();
	priority = 0.0f;
	actiontype = Action_none;
}

/*************	VisualRequestQueue	*************************/
VisualRequest::VisualRequest(float priority, AngleDeg angle){
	this->priority = priority;
	this->angle = NormalizeAngle(angle);
}

AngleDeg VisualRequest::RequestedAngle_(){
	return MidAngle(angle,angle_inf);
}

float VisualRequest::RequestedPriority(){
	if(view_width == VW_Narrow) return priority * ClientParam::VW_Narrowfactor;
	else return priority;
}

/**********		Mediator	**************************/
Mediator::Mediator(){
}

void Mediator::ResetBuffer(){
//	visualrequestqueue.cleanup();
//	countervisualrequestqueue.cleanup();
	vr_queue.cleanup();
	ac_queue.cleanup();
	totalvisualrequests = 0;
}


bool Mediator::enroll(PrioredCommand& pcmd){
	DoLog("Enroll cmdtype %d actype %d p %.2f", pcmd.type, pcmd.actiontype, pcmd.priority);
	if(pcmd.type == CMD_none) return false;
	pcmd.priority = priority_mapping(pcmd.actiontype, pcmd.priority);
	return ac_queue.enqueue(pcmd);
}

bool Mediator::enroll(Command& command, ActionType actiontype, float priority){
	PrioredCommand pricmd = PrioredCommand(command, actiontype, priority);
	return enroll(pricmd);
}

void Mediator::Getbestaction(PrioredCommand& bestaction){
	bestaction.Reset();
	if(!ac_queue.IsEmpty()){
		bestaction = ac_queue.GetData(ac_queue.ActualHead());
	}
}

void Mediator::SetViewWidth(VIEWWIDTH view_width){
	if(view_width == VW_Narrow)
		DoLog(LOG_MEDIATION, "force narrow mode");
	else if(view_width == VW_Normal)
		DoLog(LOG_MEDIATION, "force normal mode");
	else
		DoLog(LOG_MEDIATION, "force wide mode");
	designated_viewwidth.Setdata(view_width, situation.CurrentTime);
}

bool Mediator::IsViewWidthSet(){
	return bool(designated_viewwidth.time == situation.CurrentTime);
}

void Mediator::SetViewRange(){
	float bdfy = Self.bodyfacing;//bdfy when exec this command

	if(IsViewWidthSet()){
		//ѾӾģʽѾ(ȷ), ֱڵӾģʽ
		if(Self.NextViewCycle > situation.CurrentTime) {
			DoLog(LOG_SEE,"View2: %d",sensory.NextViewWidth);
			Validate_default_angle = Self.MyViewAngle(sensory.NextViewWidth); 
		}
		else{
			DoLog(LOG_SEE,"Time2: %d View2:%d",situation.CurrentTime,designated_viewwidth.data);
			Validate_default_angle = Self.MyViewAngle(designated_viewwidth.data);
		}
	}else{
		Validate_default_angle = Self.MyViewAngle(VW_Normal);
	}
	DoLog(LOG_VDEC, "view angle %.1f", Validate_default_angle);

	PrioredCommand bestaction;
	Getbestaction(bestaction);
	if (bestaction.type == CMD_turn){
		bdfy = NormalizeAngle(bestaction.angle * Skill::action.Turnrate() + bdfy);
	}
	Vang_inf = NormalizeAngle(bdfy - ServerParam::max_neck_angle - Validate_default_angle);
	Vang_sup = 2 * (ServerParam::max_neck_angle + Validate_default_angle);
	if(bestaction.type == CMD_none || bestaction.type == CMD_stay ){
		Vang_inf = NormalizeAngle(Vang_inf - Skill::action.Max_TurnAng());
		Vang_sup =  Vang_sup + 2 * Skill::action.Max_TurnAng();
	}
}

bool Mediator::ValidRequestAng(float ang, float shrinked_angle){
	return NormalizeAngle(ang - (Vang_inf + shrinked_angle) , 0) < Vang_sup - 2*shrinked_angle;
}

bool Mediator::enroll(float priority, AngleDeg angle){
	if(vr_queue.NumDatas() >= CP_max_visualrequests) return false;
	DoLog(LOG_MEDIATION, "enroll %.2f %.2f", priority, angle);
	return vr_queue.enqueue(VisualRequest(priority_mapping(Action_visual, priority), angle));
}

int Mediator::SetVisualRequests(VisualRequest * requests, int& totalrequests, VIEWWIDTH view_width, int max_requests){
	int p, q;
	float ang, priority, sum_priority, angle_inf, angle_sup, view_width_ang;
	VisualRequest treq;
	int max_idx = -1;
	float maxpriority = 0.0f;

	if (totalrequests < 0) return max_idx;
	if (vr_queue.IsEmpty()) return max_idx;

	priority = sum_priority = 0.0f;
	
	view_width_ang = Self.MyViewWidth(view_width);
	
	//ûȷӾģʽʱ, ĬϰVW_NormalĿɴǶύǶ, VW_NarrowҪ
	bool validation = bool(designated_viewwidth.time != situation.CurrentTime &&  view_width == VW_Narrow);
	float shrinked_angle = 0;
	if(validation){
		shrinked_angle = Self.MyViewAngle(VW_Normal) - Self.MyViewAngle(view_width);
		if(shrinked_angle < 0) validation = false;
	}

	p = q = vr_queue.ActualHead();
	VisualRequest* pvr;
	if(validation){
		while(vr_queue.IsValid(p)){
			pvr = &vr_queue.GetData(p);
			pvr->valid = ValidRequestAng(pvr->angle, shrinked_angle);
			p = vr_queue.Next(p);
		}
		p = vr_queue.ActualHead();
	}
	//set visual-requests clockwisely
	while(vr_queue.IsValid(p)){
		if (totalrequests >= max_requests)//exceed max requests
			break;

		pvr = &vr_queue.GetData(p);
		if(validation && !pvr->valid){
			p = vr_queue.Next(p);
			continue;
		}
		requests[totalrequests].angle = pvr->angle;
		requests[totalrequests].view_width = view_width;
		requests[totalrequests].valid = true;

		if (sum_priority > 0){
			requests[totalrequests].priority = sum_priority
				= priority_sub(sum_priority, priority);
		}
		else{
			sum_priority = requests[totalrequests].priority = 0.0f;
		}

		if(p == q){
			requests[totalrequests].angle_inf = requests[totalrequests].angle;
		}
		else{
			requests[totalrequests].angle_inf = angle_inf;
		}
		
		pvr->view_width = view_width;
		priority = pvr->RequestedPriority();

		do{
			treq = vr_queue.GetData(q);
			treq.view_width = view_width;
			ang = NormalizeAngle(treq.angle - requests[totalrequests].angle, 0.0f);
			if (ang > view_width_ang){
				break;
			}			
			if(!validation || treq.valid){
				angle_inf = requests[totalrequests].angle_inf = treq.angle;
				sum_priority = requests[totalrequests].priority = 
					priority_sum(requests[totalrequests].priority, treq.RequestedPriority());
			}
			q = vr_queue.CircularNext(q);
		}while(q != p && vr_queue.IsValid(q));
		//when searching counterclockwisely, some requests will recur, seting angle_inf
		//will help to detect it
		pvr->angle_inf = angle_inf;

		if(requests[totalrequests].priority > maxpriority){
			maxpriority = requests[totalrequests].priority ;
			max_idx = totalrequests;
		}
		totalrequests ++;
		p = vr_queue.Next(p);
	}
//set visual-requests counter-clockwisely
	p = q = vr_queue.Tail();
	priority = sum_priority = 0.0f;
	int sup_element = -1;
	
	while(!vr_queue.IsHead(p) && vr_queue.IsValid(p)){
		if (totalrequests >= max_requests)//exceed max requests
			break;

		pvr = &vr_queue.GetData(p);
		if(validation && !pvr->valid){
			p = vr_queue.Prev(p);
			continue;
		}
		requests[totalrequests].angle_inf = pvr->angle;
		requests[totalrequests].view_width = view_width;
		requests[totalrequests].valid = true;

		if (sum_priority > 0){
			requests[totalrequests].priority = sum_priority 
				= priority_sub(sum_priority, priority);
		}
		else{
			sum_priority = requests[totalrequests].priority = 0.0f;
		}

		if(p == q){
			requests[totalrequests].angle = requests[totalrequests].angle_inf;
		}
		else{
			requests[totalrequests].angle = angle_sup;
		}

		pvr->view_width = view_width;
		priority = pvr->RequestedPriority();

		do{
			treq = vr_queue.GetData(q);
			treq.view_width = view_width;
			ang = NormalizeAngle(requests[totalrequests].angle_inf - treq.angle, 0.0f);
			if (ang > view_width_ang){
				break;
			}
			if(!validation || treq.valid){
				angle_sup = requests[totalrequests].angle = treq.angle;
				sum_priority = requests[totalrequests].priority = 
					priority_sum(requests[totalrequests].priority, treq.RequestedPriority());

				sup_element = q;
			}
			q = vr_queue.CircularPrev(q);
		}while(q != p && vr_queue.IsValid(q));

		if (vr_queue.IsValid(sup_element)){
			if (vr_queue.GetData(sup_element).angle_inf != vr_queue.GetData(p).angle){				
				if(requests[totalrequests].priority > maxpriority){
					maxpriority = requests[totalrequests].priority;
					max_idx = totalrequests;
				}
				totalrequests ++;
			}
		}

		p = vr_queue.Prev(p);
	}

	return max_idx;
}

int Mediator::SetVisualRequests(){
	if (designated_viewwidth.time == situation.CurrentTime){
		DoLog(LOG_SEE,"Time: %d",situation.CurrentTime);
		//ѾӾģʽѾ(ȷ), ֱڵӾģʽ
		int max_vreq_idx = -1;
		if(Self.NextViewCycle > situation.CurrentTime){
			max_vreq_idx = SetVisualRequests(visualrequests, totalvisualrequests, sensory.NextViewWidth);
		}else{
			max_vreq_idx = SetVisualRequests(visualrequests, totalvisualrequests, designated_viewwidth.data);
		}
		sensory.SetDesireView(designated_viewwidth.data, VQ_High);

		return max_vreq_idx;
	}else{
		int vreq_idx0, vreq_idx1;
		vreq_idx0 = SetVisualRequests(visualrequests, totalvisualrequests, VW_Narrow);
		if(vreq_idx0 >= 0){
			Swap(visualrequests[0], visualrequests[vreq_idx0]);
			vreq_idx0 = 0;
			totalvisualrequests = 1;
		}
		vreq_idx1 = SetVisualRequests(visualrequests, totalvisualrequests, VW_Normal);
		if(vreq_idx0 < 0){
			sensory.SetDesireView(VW_Normal, VQ_High);
			return vreq_idx1;
		}
		if(vreq_idx1 < 0){
			DoLog("bug, shouldn't be here");
			sensory.SetDesireView(VW_Normal, VQ_High);
			return vreq_idx0;
		}
		if(visualrequests[vreq_idx0].RequestedPriority() > visualrequests[vreq_idx1].RequestedPriority()){
			sensory.SetDesireView(VW_Narrow, VQ_High);
		}else{
			sensory.SetDesireView(VW_Normal, VQ_High);
		}
		if(sensory.NextViewWidth > VW_Narrow){
			return vreq_idx1;
		}else{
			return vreq_idx0;
		}
	}
}

void Mediator::mediation(){
	
	PrioredCommand bestaction;
	Command cmd1, cmd2;
	VisualRequest tmprequest;
	float ang, bdfy,  bestaction_bdfy;
	int vreq_idx, p, max_vreq_idx;

	max_vreq_idx = SetVisualRequests();

	bestaction.Reset();

	p = ac_queue.ActualHead();
	if(ac_queue.IsValid(p)){
		bestaction = ac_queue.GetData(p);
		DoLog(LOG_ACTION,"action priority %.2f",bestaction.GetPriority());
		bdfy = Self.bodyfacing;//bdfy when exec this command
		if (bestaction.type == CMD_turn){
			bdfy = NormalizeAngle(bestaction.angle * Skill::action.Turnrate() + bdfy);
		}		
		bestaction_bdfy = bdfy;
	}

	vreq_idx = max_vreq_idx;
	submitaction = bestaction;
//execute it
	if (bestaction.type != CMD_none && bestaction.type != CMD_stay){
		Skill::action.execute(bestaction);
		if (vreq_idx != -1){
			Agent::motion.turn_neck_to(visualrequests[vreq_idx].RequestedAngle_(),
				bestaction_bdfy, cmd2);
		}
		Skill::action.execute(cmd2);
	}
	else{
		if (vreq_idx != -1){
			float visual_angle = visualrequests[vreq_idx].RequestedAngle_();
			if(bestaction.type == CMD_none || bestaction.angle > 180.0f){
				bdfy = NormalizeAngle(Self.bodyfacing +	Agent::motion.turn_to(visual_angle, cmd1));
			}else{
				RingSector vr;
				vr.Set(1.0f, 1.0f, visual_angle, ServerParam::max_neck_angle);
				vr.Intersect(NoiseData(1.0f, 1.0f), NoiseData(Self.bodyfacing, Skill::action.Max_TurnAng()));
				if(vr.IsValid()){
					if(vr.IsWithin(bestaction.angle)){
						ang = bestaction.angle; 
					}else{
						if(NormalizeAngle(vr.angles.Data(0).mean - bestaction.angle) > 0){
							ang = vr.angles.Data(0).inf();
						}else{
							ang = vr.angles.Data(0).sup();
						} 
					} 
					bdfy = NormalizeAngle(Self.bodyfacing +	Agent::motion.turn_to(ang, cmd1));
					Agent::motion.turn_neck_to(visual_angle, bdfy, cmd2); 
				}else{
					//normally it won't be here.
					ang = NormalizeAngle(visual_angle - bestaction.angle);
					pose_limitation(ang, ServerParam::min_neck_angle, ServerParam::max_neck_angle);
					ang = NormalizeAngle(visual_angle - ang);
					bdfy = NormalizeAngle(Self.bodyfacing +	Agent::motion.turn_to(ang, cmd1));
					DoLog("smart stay(dbdfc%.0f, v%.0f, bdfy%.0f)", bestaction.angle, visual_angle, bdfy);
				}
			}
			Agent::motion.turn_neck_to(visual_angle, bdfy, cmd2);
			Skill::action.execute(cmd1);
			Skill::action.execute(cmd2);
		}
	}
//feedback
	feedback(bestaction, vreq_idx);
}

void Mediator::feedback(const PrioredCommand& bestaction, int vr_idx){
	lastactiontype = bestaction.actiontype;
	lastcmdtype = bestaction.type;
	switch (bestaction.actiontype){
	case Action_interception:
		DoLog(LOG_ACTION, "Interception");
		break;
	case Action_turn_ball:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION,"Turn Ball");
		break;
	case Action_shoot:
		KickOut = TimedData<bool>(true, situation.CurrentTime);
		DoLog(LOG_ACTION,"Shoot (spd %.2f angle %.2f)", Agent::ballhandle.logshoot.MinShootSpeed(), Agent::ballhandle.logshoot.ShootAngle());
		break;
	case Action_dribble:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION,"Dribble ang %.1f",Agent::ballhandle.logdribble);
		break;
	case Action_fastdribble:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION,"FasetDribble");
		break;
	case Action_positioning_coordinate:
		DoLog(LOG_ACTION,"Coordinated attack");
		break;
	case Action_avoidenemy:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION,"Avoid enemy");
		break;
	case Action_pass:
		PassOut.time = situation.CurrentTime;
		KickOut = TimedData<bool>(PassOut.data, situation.CurrentTime);
		DoLog(LOG_ACTION,"Pass to %d(spd %.2f,angle %.1f)",
			Agent::ballhandle.logpass.Receiver(), Agent::ballhandle.logpass.PassSpeed(), Agent::ballhandle.logpass.PassAngle());
		break;
	case Action_passwait:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION,"(Waiting)Pass to %d(spd %.2f,angle %.1f)",
			Agent::ballhandle.logpass.Receiver(), Agent::ballhandle.logpass.PassSpeed(), Agent::ballhandle.logpass.PassAngle());
		break;
	case Action_holdball:
		KickOut = TimedData<bool>(false, situation.CurrentTime);
		DoLog(LOG_ACTION, "Hold ball");
		break;
	case Action_positioning_offense:
		DoLog(LOG_ACTION, "Position Offense");
		break;
	case Action_positioning_offside:
		DoLog(LOG_ACTION, "Position Offside");
		break;
	case Action_positioning_defense:
		DoLog(LOG_ACTION, "Position Defense");
		break;
	case Action_positioning_block:
		DoLog(LOG_ACTION, "Position Block");
		break;
	case Action_positioning_mark:
		DoLog(LOG_ACTION, "Position Mark");
		break;
	case Action_press:
		DoLog(LOG_ACTION, "Press");
		break;
	case Action_other:
		DoLog(LOG_ACTION, "Other");
		break;
	case Action_clear:
		KickOut = TimedData<bool>(true, situation.CurrentTime);
		DoLog(LOG_ACTION, "ClearBall");
		break;
	case Action_none:
		if (!situation.ClockStopped && situation.playmode == PM_Play_On)
			DoLog(LOG_ACTION, "None");
		break;
	default:
		DoLog(LOG_ACTION, "Not defined");
		break;
	}
	if (vr_idx != -1){
		DoLog(LOG_ACTION, "Visual request %d (%.2f %.2f)", vr_idx, visualrequests[vr_idx].angle, visualrequests[vr_idx].angle_inf);
	}
}

ActionType Mediator::LastActionType(){
	return lastactiontype;
}

CMDType Mediator::LastCmdType(){
	return lastcmdtype;
}

float Mediator::priority_mapping(ActionType actiontype, float priority){
	switch(actiontype){
	case Action_pass:
	case Action_passwait:
	case Action_shoot:
	case Action_dribble:
	case Action_fastdribble:
	case Action_catch:
	case Action_turn_ball:
	case Action_holdball:
	case Action_avoidenemy:
	case Action_clear:
		return priority * 0.2f + 0.8f;
	case Action_visual:
		return priority * 0.4f;
	case Action_interception:
		return priority * 0.4f + 0.4f;
	case Action_positioning_offense:
	case Action_positioning_offside:
	case Action_positioning_defense:
	case Action_positioning_block:
	case Action_positioning_mark:
	case Action_positioning_coordinate:
	case Action_positioning_setplay:
	case Action_press:
		return priority * 0.4f + 0.4f;
	default:
		break;
	}
	return priority;
}
